

/*
 * Baltimore: A Doom Engine
 * ------------------------
 * Release 1.01
 * Release Date Feb 5, 1998
 * Copyright Song Li, 1996, 1997, 1998
 * 
 * You are encouraged to use source code of Baltimore for non-commercial
 * applications. That means, you can study it, improve it, embed it in
 * your own programs, or even distribute it across the internet. Both
 * source code and binaries are free. The only condition is you are required
 * to give credits to me, Song Li, the only author of Baltimore. Please
 * keep these copyright informations while making modifications.
 *
 * Baltimore Homepage:  http://www.cs.umbc.edu/~sli2/baltimore
 * Author's Homepage:   http://www.cs.umbc.edu/~sli2
 * Author's E-Mail:     sli2@cs.umbc.edu
 *
 */


#include "Copyright.h"

#ifdef _WINDOWS
#include <windows.h>
#endif

#ifdef _UNIX
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#endif


#include "BalDef.h"
#include "Painter.h"
#include "Buffer.h"
#include "Map.h"
#include "Camera.h"
#include "Player.h"
#include "Rengine.h"
#include "Timer.h"
#include "File.h"




const unsigned int DEMOWIDTH     = 320;
const unsigned int DEMOHEIGHT    = 200;
const BREAL        DEMODIST      = 120.0f;
const BREAL        DEMOPROJECTOR = 0.012f;


static BBuffer   * vbuffer = BNULL;
static BPainter  * painter = BNULL;
static BRengine  * engine  = BNULL;
static BMap      * map     = BNULL;
static BPlayer   * player  = BNULL;
static BTimer    * timer   = BNULL;



BBOOL keyup    = BFALSE;
BBOOL keydown  = BFALSE;
BBOOL keyleft  = BFALSE;
BBOOL keyright = BFALSE;






#ifdef _WINDOWS
 

#define BM_REFRESH   (WM_USER+1)





static void BalRender (HWND hWnd, HDC hdc) {
    static int time = 0;
    static int repeat = 1000;
    BTimer::Execute (time++);


    if (keyup)
        player->Forward  (1.3f);
    if (keydown)
        player->Backward (1.3f);
    if (keyleft)
        player->Turn (-BPI/30.0f);
    if (keyright)
        player->Turn ( BPI/30.0f);


    BCamera * camera = new BCamera (* player);
    engine->Render (* vbuffer, * camera);
    painter->Paint (hdc, 0, 0);
    delete camera;
    
    PostMessage (hWnd, BM_REFRESH, 0, 0);
    return;
}






long FAR PASCAL BalProc (HWND hWnd, UINT message, 
                         WPARAM wParam, LPARAM lParam) {

    PAINTSTRUCT ps;
    HDC hdc;


    switch (message) {
    case WM_CREATE:
         break;


    case WM_KEYDOWN:
    case WM_KEYUP:
         switch (wParam) {
             case VK_ESCAPE:
             case VK_F12:
                  PostMessage (hWnd, WM_CLOSE, 0, 0);
                  break;

             case VK_LEFT:  keyleft  = (message==WM_KEYDOWN); break;
             case VK_RIGHT: keyright = (message==WM_KEYDOWN); break;
             case VK_UP:    keyup    = (message==WM_KEYDOWN); break;
             case VK_DOWN:  keydown  = (message==WM_KEYDOWN); break;
             }
         
         if (keyleft || keyright || keyup || keydown)
             InvalidateRect (hWnd, NULL, BFALSE);
         break;


    case WM_PAINT:
         hdc = BeginPaint (hWnd, & ps);
         BalRender (hWnd, hdc);
         EndPaint (hWnd, & ps);
         break;
    

    case BM_REFRESH:
         InvalidateRect (hWnd, NULL, FALSE);
         break;

    case WM_DESTROY:
         PostQuitMessage( 0 );
         break;
    }


    return DefWindowProc (hWnd, message, wParam, lParam);
}




static BOOL BalInit (HINSTANCE hInstance, int nCmdShow) {
    timer   = new BTimer   ();
    engine  = new BRengine (DEMOWIDTH, DEMOHEIGHT, DEMODIST, DEMOPROJECTOR);
    map     = new BMap     ();
    player  = new BPlayer  (BPoint   (40.0f, 10.0f),
                            BVector  ( 0.0f,  1.0f),
                            5.0f,
                            map->Sector (0));
    vbuffer = new BBuffer  (DEMOWIDTH, DEMOHEIGHT);
    painter = new BPainter (* vbuffer);


    WNDCLASS  wc;

    wc.style         = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc   = BalProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = 0;
    wc.hInstance     = hInstance;
    wc.hIcon         = LoadIcon (hInstance, IDI_APPLICATION);
    wc.hCursor       = LoadCursor (NULL, IDC_ARROW);
    wc.hbrBackground = GetStockObject (LTGRAY_BRUSH);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = "Baltimore 3D";
    RegisterClass (& wc);
    

    HWND hwnd = CreateWindow (
        "Baltimore 3D",
        "Baltimore 3D",
        WS_POPUPWINDOW | WS_CAPTION,
        50, 200, 
        DEMOWIDTH, DEMOHEIGHT + GetSystemMetrics (SM_CYCAPTION),
        NULL, NULL, hInstance, NULL);

         
    ShowWindow (hwnd, nCmdShow);
    UpdateWindow (hwnd);
    return TRUE;
}




static void BalDeinit () {
    delete painter;
    delete vbuffer;
    delete player;
    delete map;
    delete engine;
    delete timer;
}



int PASCAL WinMain (HINSTANCE hInst, HINSTANCE, LPSTR, int nCmdShow) {
    MSG msg;

    if (BalInit (hInst, nCmdShow) == FALSE)
        return FALSE;


    while (GetMessage (& msg, NULL, 0, 0)) {
        TranslateMessage (& msg);
        DispatchMessage  (& msg);
        }

    BalDeinit ();
    return msg.wParam;
}


#endif






#ifdef  _UNIX



static Display * display;
static int       screen;
static Visual  * visual;
static Window    window;
static GC        gc;







Colormap BalPalette () {
    BBYTE pal [256 * 3];
    BFile file;
    file.Open  ("baltimore.pal");
    file.Read  (pal, 256 * 3);
    file.Close ();


    Colormap colormap;
    colormap = XCreateColormap (display, 
                                XRootWindow (display, screen), 
                                visual, AllocNone);

    for (int i=0; i<256; i++) {
        XColor color;
        color.pixel = i;
        color.red   = 256 * pal[i*3];
        color.green = 256 * pal[i*3+1];
        color.blue  = 256 * pal[i*3+2];
        XAllocColor (display, colormap, & color);
        }

    return colormap;
}






static BBOOL BalInit ()  {

    display = XOpenDisplay (NULL);
    if (display == NULL) 
        return BFALSE;
    screen = DefaultScreen (display);



    XVisualInfo vtemplate, * vinfo;
    int vcount;
    vtemplate.screen  = screen;
    vtemplate.c_class = PseudoColor;
    vtemplate.depth   = 8;
    vinfo = XGetVisualInfo (display, 
                 VisualScreenMask | VisualClassMask | VisualDepthMask,
                 & vtemplate, & vcount);
    if (vinfo == NULL)
        return BFALSE;
    visual = vinfo->visual;
    XFree (vinfo);
    

    XSetWindowAttributes winattr;
    winattr.colormap     = BalPalette ();
    winattr.border_pixel = 0;
    winattr.event_mask   = ExposureMask | KeyReleaseMask | KeyPressMask;
    window = XCreateWindow (display, XRootWindow (display, screen),
                            200, 200, DEMOWIDTH, DEMOHEIGHT, 0,
                            8, InputOutput, visual,
                            CWColormap | CWBorderPixel | CWEventMask, 
                            & winattr);

    gc = XCreateGC (display, window, 0, NULL);
    


    timer   = new BTimer   ();
    engine  = new BRengine (DEMOWIDTH, DEMOHEIGHT, DEMODIST, DEMOPROJECTOR);
    map     = new BMap     ();
    player  = new BPlayer  (BPoint   (40.0f, 10.0f),
                            BVector  ( 0.0f,  1.0f),
                            5.0f,
                            map->Sector (0));
    vbuffer = new BBuffer  (DEMOWIDTH, DEMOHEIGHT);
    painter = new BPainter (* vbuffer, display);

    
    XFlush (display);
    XMapWindow (display, window);
    XFlush (display);


    return BTRUE;
}




static void BalDeinit () {
    delete painter;
    delete vbuffer;
    delete player;
    delete map;
    delete engine;
    delete timer;

    XDestroyWindow (display, window);
    XCloseDisplay  (display);
}

    




static void BalRender (...) {
    static int time = 0;
    BTimer::Execute (time++);
    
    if (keyup)
        player->Forward  (1.3f);
    if (keydown)
        player->Backward (1.3f);
    if (keyleft)
        player->Turn (-BPI/30.0f);
    if (keyright)
        player->Turn ( BPI/30.0f);


    BCamera * camera = new BCamera (* player);
    engine->Render (* vbuffer, * camera);
    painter->Paint (window, gc, 0, 0);
    delete camera;


    XExposeEvent event;
    event.type = Expose;
    event.serial = 0;
    event.send_event = True;
    event.display = display;
    event.window  = window;
    event.x = 0;
    event.y = 0;
    event.width  = DEMOWIDTH;
    event.height = DEMOHEIGHT;
    event.count  = 1;
    XSendEvent (display, window, False, ExposureMask, (XEvent *) & event);
    return;
}





static void BalProc () {
    XEvent  event;
    KeySym  keysym;
    BBOOL   quit = BFALSE;


    do {
        XNextEvent (display, & event);

        switch (event.type) {
            case Expose:
                 BalRender ();
                 break;

            case KeyPress:
            case KeyRelease:
                 keysym = XLookupKeysym ((XKeyEvent *) & event, 0);
                 switch (keysym) {
                     case XK_Escape:
                          quit = BTRUE;
                          break;

                     case XK_KP_Up:
                     case XK_Up:
                          keyup = (event.type==KeyPress);
                          break;
                     
                     case XK_KP_Down:
                     case XK_Down:
                          keydown = (event.type==KeyPress);
                          break;
                     
                     case XK_KP_Left:
                     case XK_Left:
                          keyleft = (event.type==KeyPress);
                          break;
                     
                     case XK_KP_Right:
                     case XK_Right:
                          keyright = (event.type==KeyPress);
                          break;
                     }
                 break;
            }
    } while (quit == BFALSE);
}





int main () {
    
    if (BalInit () == BFALSE) {
        BMessage ("Cannot initialize X routines.");
        return 1;
        }
    
    BalProc ();
    
    BalDeinit ();
    return 1;
}


#endif

